; LAML, LENO, and ELUCIDATOR template support and other useful editing commands on LAML files.

; The LAML library and programs written by Kurt Normark, Aalborg University, Denmark.
; Copyright (C) 1999  Kurt Normark, normark@cs.auc.dk.
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; Top level functions in this package
; insert-f for each LENO function f. 
; Also insert manual page/section, and elucidator templates.
; The functions embed, nest and unnest are also very useful (in general).

; Notice: note-page templates are supported via the general template facily,
; which is now part of this file.


; This variable depends on the variable laml-scheme-dir, which is supposed to be defined
; in the emacs init file (emacs.init or .emacs)

(defvar laml-elisp-dir (concat laml-dir "emacs-support/") "The directory which contains this file. Ends in a slash")

(defvar laml-pp-helpful t 
  "Show useful role names of constituents in selected LAML templates. 

If t always pretty print role strings in templates. If nil, just print empty strings for some template elements
LAML templates are activated by M-x insert-... in the various LAML styles. In LENO, use M-x leno-insert-...")



; ---------------------------------------------------------------------------------------------------
; General template support for emacs:


(defvar laml-template-dir (concat laml-elisp-dir "templates")
   "The directory path to the
directory, in which the templates (tpl as well as el) files are stored.")

(defun laml-insert-template (tmpl-name)
  "Inserts a template from the template collection in the directory
templates."
  (interactive 
    (list
      (completing-read
       "Template name: "
       (mapcar 'laml-tmpl-transform-name
               (filter 'laml-tmpl-tpl-filter
                       (file-name-all-completions "" laml-template-dir)))
             nil t nil)))
  (insert-file (concat laml-template-dir "/" tmpl-name ".tpl"))
  (let ((el-file (concat laml-template-dir "/" tmpl-name ".el")))
  (if (file-exists-p el-file)
      (load-file el-file))))

(defun laml-get-template (tmpl-name)
  "Return the template string of tmpl-name"
  (let ((tmpl-path (concat laml-template-dir "/" tmpl-name ".tpl")))
    (read-text-file tmpl-path)))  ; not implemented


(defun insert-laml-template ()
    "Return a template from the template collection in the directory
templates. Returns a string"
  (interactive)
  (call-interactively 'laml-insert-template))

;; Functions that help generate a completion list of template names:

(defun laml-tmpl-transform-name (n)
  "Function used for the production of the template name completion list"
  (cons (substring n 0 (- (length n) 4)) nil))

(defun laml-tmpl-tpl-filter (n)
  (equal (substring n (- (length n) 4)) ".tpl"))


; ---------------------------------------------------------------------------------------------------
; Template insertion
  
(defun insert-lisp-form (form indentation)
  (setq indent-tabs-mode nil)
  (ensure-indentation 0)
  (insert (concat (make-string indentation 32) form))
  (backward-sexp 1)
  (indent-sexp)
  (forward-sexp 1) (beginning-of-line)
  (ensure-indentation indentation)
  (end-of-line)
  (backward-sexp 1))

; old, but still used
(defun insert-lisp-form-0 (form indentation)
  (insert-lisp-form (laml-transliterate form) indentation))


; ---------------------------------------------------------------------------------------------------
; Transliteration stuff:

(defvar empty-string "\"\"")

(defun laml-transliterate (form)
  "Return a copy with various transliterations" 
  (let* ((res1 (string-to-list form))
         (res2 (if (eq (car res1) 40)
                   (cons 40 (transliterate-list (cdr res1) 40 (concat CR "(")))  ; avoid CR before first start parenthesis
                   (transliterate-list res1 40 (concat CR "("))))
         (res3 (transliterate-list res2 41 (concat ")" )))
         (res4 (transliterate-list res3 ?E (concat empty-string CR )))
        )
    (list-to-string res4)))


(defun transliterate-list (list char x)
  "List is a list of characters (integers). Replaces occurences of char with x.
The parameter x can be a string or a character"
  (let ((list1 list)
        (res nil))
    (while list1 
      (setq res
        (cons 
          (if (eq (car list1) char)
              x
              (car list1))
          res))
      (setq list1 (cdr list1)))
    (reverse res)))


(defun charlist-to-string (lst)
  "The opposite of string-to-list"
  (let ((str (make-string (length lst) 0))
        (lst1 lst)
        (i 0))
    (while lst1 
       (aset str i (car lst1))
       (setq i (+ i 1))
       (setq lst1 (cdr lst1)))
    str))

(defun list-to-string (lst)
  "Convert a list of mixed chars and strings to a string"
  (let ((str (make-string (length-char-string-list lst) 0))
        (lst1 lst)
        (i 0))
    (while lst1 
       (aset-char-string str i (car lst1))
       (setq i (+ i (length-char-string (car lst1))))
       (setq lst1 (cdr lst1)))
    str))

(defun length-char-string-list (lst)
  "Calculate the length of lst. Chars count 1, strings count their lengths"
 (let ((res 0)
       (lst1 lst))
  (while lst1
    (setq res (+ res (length-char-string (car lst1))))
    (setq lst1 (cdr lst1)))
  res))

(defun length-char-string (x)
  "Return the length of x. Chars count 1, strings count their lengths"
  (cond ((integerp x) 1)
        ((stringp x) (length x))))

(defun aset-char-string (str i x)
  "like aset, but x may also be a string, in which case all chars in x are inserted"
  (cond ((integerp x) (aset str i x))
        ((stringp x) (aset-string str i x 0))))

(defun aset-string (str i str1 j)
  "Assign destructively str1 into position i in str. j is a helping, indexing parameter"
  (if (= j (length str1))
      nil  ; result in str
      (progn 
        (aset str i (aref str1 j))
        (aset-string str (+ i 1) str1 (+ j 1)))))
          


; ---------------------------------------------------------------------------------------------------
; Indentation stuff:

(defun determine-indentation ()
  (let ((ci (real-indentation-of-string (this-line))))
    (if (> ci 0)
        ci
        (indentation-of-previous-line))))

   
(defun indentation-of-previous-line ()
  "Return the indentation of the previous line. Here previous line means the previous
real line. Empty lines are not taken into consideration. In case no previous line exist, return 0"
  (save-excursion
    (beginning-of-line)
    (let ((prev (prev-line)))
      (if prev
          (cond ((space-stringp prev) (progn (previous-line 1) (indentation-of-previous-line)))
                (t (indentation-of-string prev)))
          0))))

(defun real-indentation-of-string (str)
  (cond ((equal str "") 0)
	((space-stringp str) (length str))
	(t (let ((i 0))
	     (while (eq (aref str i) 32)
	       (setq i (+ i 1)))
	     i))))

(defun indentation-of-string (str)
  "Return the number of leadning spaces of str. 
If string is empty, or if it consists soly of spaces, return 0"
 (if (or (equal str "") (space-stringp str))
     0
     (let ((i 0))
       (while (eq (aref str i) 32)
          (setq i (+ i 1)))
       i)))

(defun space-stringp (str)
  "Is str full of spaces"
  (or (= 0 (length str)) 
      (equal str (make-string (length str) 32))
  ))

(defun prev-line ()
 (save-excursion
  (let ((p1 (point)))
    (previous-line 1)
    (if (eq p1 (point))
        nil
        (this-line)))))

(defun this-line ()
  (save-excursion 
    (beginning-of-line)
    (let ((p1 (point)))
      (end-of-line)
      (buffer-substring p1 (point)))))    
     
(defun ensure-indentation (i)
  "Enforce that the current line is idented i positions."
  (save-excursion
   (let ((ci (real-indentation-of-string (this-line))))
      (if (not (= i ci))
          (progn
             (beginning-of-line)
             (delete-char ci)
             (insert-char 32 i))))))


; ---------------------------------------------------------------------------------------------------
; File making from template.
; (Originally located in elucidator.el)

(defun make-a-file-from-laml-template (file dir template-name emacs-mode replacements)
  "Make the FILE in DIR. Then insert the LAML template TEMPLATE-NAME. 
Finally carry out the list of REPLACEMENTS. Return the files buffer"
  (let ((new-buffer (create-file-buffer (concat dir file))))
    (set-buffer new-buffer)
    (eval (list emacs-mode)) ; set the mode
    (laml-insert-template template-name)
    (do-perform-replacements replacements)
    (write-file (concat dir file))
    new-buffer))

(defun do-perform-replacements (replacement-list)
  (mapcar
    (function do-perform-replacement)
    replacement-list))

(defun do-perform-replacement (repl)
  (let ((from-string (car repl))
        (to-string (cadr repl)))
    (goto-char (point-min))
    (while (search-forward from-string nil t)
      (replace-match to-string t t))))


(defun ensure-trailing-slash (dir-string)
  "Ensure that dir-string ends in a slash: /"
  (let* ((dir-lgt (length dir-string))
         (last-chr (aref dir-string (- dir-lgt 1))))
   (if (eq last-chr ?/)
       dir-string
       (concat dir-string "/"))))
  
  
; move to general.el (also in making-directories).
(defun determine-directory ()
  "What is the current directory"
  (cond ((buffer-file-name) (current-directory))
        ((eq major-mode 'dired-mode) (dired-current-directory))
        (t nil)))
             



; ---------------------------------------------------------------------------------------------------

; String splitting stuff:

(defvar QUOTE-CHAR 34)
(defvar SPACE-CHAR 32)
(defvar QUOTE-STRING (char-to-string QUOTE-CHAR))

(defun goto-start-of-string ()
  "Assumes that the point is inside a string (not point at a quote).
Move the point to the start of the string, pointing at the quote character"
  (while (not (eq (char-under-point) QUOTE-CHAR))
         (backward-char 1)))

(defun goto-end-of-string ()
  "Assumes that the point is inside a string (not point at a quote).
Move the point to the end of the string, poining at the quote character"
  (while (not (eq (char-under-point) QUOTE-CHAR))
         (forward-char 1)))

(defun char-under-point (&optional n)
  "Return the character under point. If n is supplied, it must be a number.
   In that case n is added to point before the character is returned. In that
   way we can access chacters before (-1) or after (1) point too."
  (let* ((n1 (if n n 0))
         (pt1 (+ (point) n1)))
    (string-to-char (buffer-substring pt1 (+ 1 pt1)))))

(defun word-delimitor-scheme ()
  "Return a list of two numbers: the start and the end positions of the Scheme name under point"
 (save-excursion
  (let ((rb (progn (goto-start-of-scheme-word)
                 (point)))
        (re (progn (goto-end-of-scheme-word)
                 (point))))
    (if (and rb re) (list rb re)))))

(defun scheme-name-under-point ()
  (let ((wd (word-delimitor-scheme)))
    (buffer-substring (car wd) (+ 1  (cadr wd)))))

; remove all spaces in the neighborhood of the location p
; Return the number of spaces removed
(defun space-remove (p)
 (let ((i 0))
  (save-excursion 
   (goto-char p)
   ; first go back without deleting spaces:
   (if (looking-at " ") (progn (while (looking-at " ") (backward-char 1)) (forward-char 1)))

   ; Now delete and count in the variable i:
   (while (looking-at " ") (progn (delete-char 1) (setq i (+ i 1)))))
  i))


(defun split-string-into-three ()
  "Assume that point is within a string and that point and mark defines a selection. Split the current string into
   three strings: The section before the selection, the selection, and the section after. Leave point at the newly inserted start
   quote of the selected string"
 (let ((qs (char-to-string QUOTE-CHAR)))
  (let ((p1 (min (point) (mark)))
        (p2 (max (point) (mark)))
        (cur-p (point)))
    
    ; remove a space to the right
    (if (not laml-single-string-approach) (space-remove p2))

    ; remove a space to the left
    (if (not laml-single-string-approach)
        (if (not (at-indentation p1))
          (let ((count (space-remove (- p1 1))))
	    (setq p1 (- p1 count)) (setq p2 (- p2 count)))))
    
    (goto-char p2) (insert (concat qs " " qs))
    (goto-char p1) (insert (concat qs " " qs))
    (goto-char (+ 2 p1))
)))


(defun at-indentation (p)
 "Is possition p an indentation possition."
 (save-excursion
  (goto-char p)
  (backward-char 1)
  (if (looking-at-chars (point) (list 32 9))
      (progn
        (while (looking-at-chars (point) (list 32 9)) (backward-char 1))
        (looking-at-chars (point) (list 13 10)))
      nil)))


(defun split-string-into-two ()
  "Assume that point is within a string. Split the current string into
   two strings: The section before the selection and the section after. Leave point between the two strings."
 (let ((qs (char-to-string QUOTE-CHAR)))
  (if (not laml-single-string-approach) (space-remove (point)))
  (let* ((p1 (point))
        )

    (goto-char p1) (insert (concat qs))
    (goto-char p1) (insert (concat qs " "))
)))


; ---------------------------------------------------------------------------------------------------------
; Generally useful LAML commands

(defvar previous-nesting "" "The previously typed nesting name - for convenient repetition of nestings.")

(defun nest (function-name)
  "Nest the form designated by the point into a call of function-name. Function-name is a symbol.
When called interactively you are prompted for a symbol in the minibuffer"
  (interactive 
    (list (string-to-symbol (completing-read "Nest into which form: " 
                               nil
                               nil nil previous-nesting))))
  (insert "(") 
  (insert (symbol-to-string function-name))
  (insert " ")
  (forward-sexp 1)
  (insert ")")
  (backward-sexp 1)
  (setq previous-nesting (symbol-to-string function-name))
)

(defun unnest ()
  "Unnest the form designated by the point. Point must be at a start or end parenthesis character.
   Leaves point just before the first constituent of the unnested form."
  (interactive)
  (if (= (char-under-point) 41)  ; end parenthesis
      (backward-up-list 1))
  (if (= (char-under-point) 40)  ; start parenthesis
      (let ((p1 (point)))
        (forward-sexp 1)
        (delete-backward-char 1)  ; delete end parenthesis
        (goto-char p1)
        (delete-char 1)  ; delete start parenthesis
        (kill-sexp 1)    ; delete function position
        (if (looking-at " ") (delete-char 1))
        
      )
  )
)


(defvar concatenate-form 'con 
  "The Scheme string concatenate function to be used with LAML. Only relevant for text-based mirrors (the very old mirrors).

The very old mirrors are simple and simple-html4.0-loose. With use of all other mirrors, just ignore this variable.
")

; currently an adhoc collection - outdated.
(defvar html-laml-embed-tag-list nil "A list of html and laml tag names (symbols) in which embedding a text string makes sense")
(setq html-laml-embed-tag-list '(em a-tag a-tag-target h kbd img strong pre center cite blockquote i b u p frame box color-frame frame-width))

;(defvar laml-single-string-approach nil 
;  "A boolean variable which defines whether we work with LAML forms which expect a single string as argument.
;The alternative is functions which accept an arbitrary list of strings as arguments.")

(defvar smart-word-determination t 
  "With smart word determination a word is implicitely selected if the point is at the first character of a word.

If you do not use smart word determination you must explicitly select a string to embed. The variable is important for the embed command.")


(defun ensure-concatenation-context ()
  "Point is assumed to be within a string. Make sure that the string is within a direct concatenation context.
If necessary, insert a con form. Con is an alias of string-append in LAML."
  (save-excursion
    (let ((p-cur (point))
          (p-stringstart nil))
      (goto-start-of-string)
      (setq p-stringstart (point))
      (backward-up-list 1) (forward-char 1) 
      (let ((form-name (string-to-symbol (scheme-name-under-point))))  ; only part before dash!
        (if (not (concatenate-formp form-name))
            (progn
               (goto-char p-stringstart)
               (nest concatenate-form)
               )))))
)


(defun concatenate-formp (x)
  (memq x (list concatenate-form)))



(defun word-interpretation ()
  "Return non-nil, if the context signals that the word under point is to be the selection.
Returns true if points happens to be located at the first character in a word"
 (save-excursion
  (backward-char 1) (or (looking-at CR) (looking-at " ") (looking-at "\""))))

(defvar previously-typed-tag "" "The previously typed tag - for reusability.")

(defun embed (tag)
  "Embed the currenly selected string/word into tag (a function call). Ensures automatically that
the function call occurs within a con concatenation context (if laml-single-string-approach is true).
If smart-word-determination is true, a word under point is selected in case point is located at the word's first letter.
The parameter tag must be a symbol"
  (interactive 
    (list (string-to-symbol (completing-read "Embed in which Scheme function: " 
                               (trivial-alist (mapcar (function symbol-to-string) html-laml-embed-tag-list))
                               nil nil previously-typed-tag))))
  (if (and smart-word-determination (word-interpretation)) (mark-sexp 1))    ; earlier: (makr-word 1)
  (if (looking-at "\"") (exchange-point-and-mark)) ; a trick to avoid problems with a embedding of the rear end of a string.
  (if laml-single-string-approach (ensure-concatenation-context))
  (split-string-into-three)
  (nest tag)
  (remove-empty-siblings)
  (fix-pending-punctuation)
  (if (mark-marker) (move-marker (mark-marker) (point)))
  (setq previously-typed-tag (symbol-to-string tag))
)

(defun fix-pending-punctuation ()
 "A specialized function which fixes the problem of a punctuation character after an embedding."
 (save-excursion
  (forward-sexp 1)
  (while (blank-char-p (char-under-point))
                (forward-char 1))

  (if (looking-at "\"")
      (progn 
        (forward-char 1)
        (if (looking-at-punctuation-char (point))
            (progn
              (forward-char 1) 
              (if (looking-at-white-space (point))
                  (progn (split) (backward-sexp 1) (insert "_ "))
                  (progn (backward-char 2) (insert "_ ")))))))))



(defun looking-at-punctuation-char (p)
 (let ((ch (char-after p)))
  (or (= ch ?.) (= ch ?,) (= ch ??) (= ch ?!) (= ch ?:)
      (= ch 59))))  ; 59 is semicolon

(defun looking-at-white-space (p)
 (let ((ch (char-after p)))
  (member ch (list 32 10 13))))

(defun looking-at-chars (p char-list)
  "is char at possition p member of char-list"
 (let ((ch (char-after p)))
  (member ch char-list)))  

(defun split ()
  "Split the string under point in two parts, and ensure (if laml-single-string-approach) that the strings parts are in a 
   concatenation context. Leave point in between the two string parts"
  (interactive)
  (if (looking-at "\"") (exchange-point-and-mark)) ; a trick to avoid problems with a embedding of the rear end of a string.
  (if laml-single-string-approach (ensure-concatenation-context))
  (split-string-into-two)
  (remove-empty-siblings)
)

; Assume as a precondition that we are inside a quoted string.
; Move point to the start quote of the string
(defun to-front-of-string ()
 (let ((cur-point (point)))
   (while (not (looking-at "\"")) (backward-char 1))))

(defun unsplit ()
  "Join neighbor verbatim strings together.
   You are supposed to select an inner character in one of the strings to be joined,
   or the start quote of one the strings.
   We assume that there are at least one space between verbatim strings. If not, this command will not work."
  (interactive)

  (save-excursion 

    (to-front-of-string)
    (forward-char -1)

    (while (and (string-to-the-left-p) (string-to-the-right-p))
           (forward-sexp -1) (forward-char -1))


    (save-excursion 
     (let* ((gaps (string-gaps (point)))
           (rev-gaps (reverse gaps)))

      (while (not (null rev-gaps))
        (progn
          (goto-char (car rev-gaps))
          (if (and (string-to-the-left-p) (string-to-the-right-p))
              (join-neighbor-strings))
          (setq rev-gaps (cdr rev-gaps))))))

    (backward-up-list 1)

    ; unnest if we see a con with a single string argument
    (if (and laml-single-string-approach (single-con-form-p (form-under-point)))
        (progn
          (unnest)
          (if (looking-at " ") (delete-char 1))
        )
    )

))
     


; Assume that possition p is at an initial quote in a string.
; Return a list of positions of gaps in between verbatim strings in front of us.
(defun string-gaps (p)
  (save-excursion (string-gaps-1 p)))

(defun string-gaps-1 (p)
  (forward-sexp 1)
  (if (and (string-to-the-left-p) (string-to-the-right-p))
      (cons (point) (string-gaps-1 (point)))
      nil))



; Is there a string to the left of point.

(defun string-to-the-left-p ()
  (save-excursion
    (while (blank-char-p (char-under-point))
      (forward-char -1))
    (equal QUOTE-CHAR (char-under-point))))


; Is there a string to the right of point.
(defun string-to-the-right-p ()
  (save-excursion
    (while (blank-char-p (char-under-point))
      (forward-char +1))
    (equal QUOTE-CHAR (char-under-point))))


;
(defun in-between-strings ()
 (save-excursion
  (cond ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point -1) QUOTE-CHAR))  ; at quote and also quote to the left
           t)
        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point 1) QUOTE-CHAR))   ; at quote and also quote to the left
           t)
        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point -1) SPACE-CHAR))
          (save-excursion
            (forward-char -1)
            (string-to-the-left-p)))
        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point 1) SPACE-CHAR))
          (save-excursion
            (forward-char 1)
            (string-to-the-right-p)))
        ((and (string-to-the-left-p) (string-to-the-right-p))
            t)
        (t nil))))



; assume we are in between to verbatim strings or at the leftmost string quote. Join them.
; Precondition: Point must be at white space in between strings, or one of the string quotes
(defun join-neighbor-strings ()

  (cond ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point -1) QUOTE-CHAR)) ; looking at " and a string quote to the left
           (delete-char 1) (delete-char -1)
           (if (not laml-single-string-approach)   ; always?
		(insert " "))
        )

        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point 1) QUOTE-CHAR)) ; looking at " and a string quote to the right
           (delete-char 2)
           (if (not laml-single-string-approach)   ; always?
		(insert " "))
        )

        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point -1) SPACE-CHAR)) ; looking at " and a space to the left of it
           (delete-char 1)  ; delete right quote
           (while (not (equal QUOTE-CHAR (char-under-point -1)))
		  (backward-delete-char-untabify 1))
           (backward-delete-char-untabify 1) ; delete left quote
           (if (not laml-single-string-approach)   ; always?
		(insert " "))
        )

        ((and (equal (char-under-point) QUOTE-CHAR) (equal (char-under-point 1) SPACE-CHAR)) ; looking at " and a space to the right of it
           (delete-char 1)  ; delete right quote
           (while (not (equal QUOTE-CHAR (char-under-point)))
		  (delete-char 1))
           (delete-char 1) ; delete left quote
           (if (not laml-single-string-approach)   ; always?
		(insert " "))
        )

        ((looking-at " ")
          (progn

            ; first join to the left:
	    (if (not (equal (char-under-point) QUOTE-CHAR))
		(while (not (equal QUOTE-CHAR (char-under-point -1)))
		  (backward-delete-char-untabify 1)))
	    (backward-delete-char-untabify 1) ; delete left string quote

	    ; next join to the right:
	    (while (not (equal QUOTE-CHAR (char-under-point 1)))
	      (delete-char 1))
	    (delete-char 2)		; delete right quote string and char in front of it.

	    (if (not laml-single-string-approach)
		(insert " "))))))



(defun single-con-form-p (x)
  (and (listp x)
       (= (length x) 2)
       (eq 'con (car x))
       (stringp (cadr x))))

; Return the lisp form under and before point. If not possible, return nil.
(defun form-under-point ()
 (save-excursion
   (let ((p (point)))
     (forward-sexp 1)
     (let ((q (point)))
       (if (and p q)
           (let ((form  (car (read-from-string (buffer-substring p q)))))
             form)
           nil)))))

(defun unembed ()
  "Unnest the currently selected form and join neighboring verbatim strings together.
   You are supposed to select a character of the form prefix you want to get rid of, or the start parenthesis."
  (interactive)

  (save-excursion  
    ;(if (not (looking-at "(")) (backward-up-list 1))
    (if (looking-at "(")
        (progn
          (unnest)
          ; now in between possible verbatim strings
      
      
          ; try to join verbatim strings to the left
          (if (in-between-strings) ; (and (string-to-the-left-p) (string-to-the-right-p))
              (join-neighbor-strings)
              (forward-char 1))
      
          ; go to end of possibly joined strings
          (while (not (equal QUOTE-CHAR (char-under-point)))
             (forward-char 1))
          (forward-char 1)
        
          ; try to join verbatim strings to the right
          (if (in-between-strings) ; (and (string-to-the-left-p) (string-to-the-right-p))
              (join-neighbor-strings)
              (forward-char -2))
      
          ; go to beginning of possibly joined strings
          (while (not (equal QUOTE-CHAR (char-under-point)))
             (forward-char -1))
          (forward-char -1)
      
          (backward-up-list 1)
      
          ; unnest if we see a con with a single string argument
          (if (and laml-single-string-approach (single-con-form-p (form-under-point)))
              (progn
                (unnest)
                (if (looking-at " ") (delete-char 1))
              )
          ))
        (message "To unembed select start parenthesis of form which you want to unembed"))))


; Given some position in a list, return the sibling constituents which are literal empty lists.
; Do not change the position of point.
(defun remove-empty-siblings ()
  (interactive)
  (let ((empty-string-reg-exp "\"\""))
   (save-excursion
    (backward-up-list 1) (forward-char 1)
    (while (not (looking-at ")"))
      (progn
        (if (looking-at empty-string-reg-exp)
            (progn
              (kill-sexp 1)
              (while (blank-char-p (char-under-point))
                (delete-char 1))
            )
            (forward-sexp 1))
        (forward-through-white-space))))))

(defun forward-through-white-space ()
  (skip-through (list 32 10 13)))  ; also include tab


; -----------------------------------------------------------------------------

; Manual templates:

(defun insert-manual-page (form-name)
  (interactive "sName of manual entry: ")
  (insert-lisp-form-0 (make-manual-page form-name) (determine-indentation))

  ; replace the QBXYZ hack: 
  (while (search-forward "(QBXYZ"  nil t)
    (replace-match  "'(" nil t))
  (beginning-of-defun))

(defun make-manual-page (form-name)
  (concat "(manual-page " CR "'" form-name 
     "(title " (string-it form-name) ")"
     "(form "  "(QBXYZ" form-name "))"  ; QBXYZ is a hack. It will be replaced by a quote and a start parenthesis
     "(description " empty-string ")"
     "(parameters"
         "(parameter " empty-string " " empty-string ")"
         "(parameter " empty-string " " empty-string ")" ")"
     "(pre-condition " empty-string ")"
     "(cross-references"
         "(reference " empty-string " " empty-string " " empty-string ")" ")"
     "(misc " empty-string ")" ")"
  ))

(defun insert-manual-section ()
  (interactive)
  (insert-lisp-form-0 (make-manual-section) (determine-indentation)))

(defun make-manual-section ()
  (concat "(manual-section" CR
     "(section-title " empty-string ")"
     "(section-body " empty-string "))"))


(defun insert-manual-internal-tag (kind)
  (interactive  (list (string-to-symbol (completing-read "Insert which kind of internal tag: " 
                      (trivial-alist "title" "form" "pre-condition" "parameter" "example" 
                                      "reference" "internal-references" "misc" "comment" "returns") nil nil)))
    kind))





; --------------------------------------------------------------

;; Replace <, > and & with character entities that present these characters properly in HTML.
;; An interactive emacs lisp counterpart to the Scheme function html-protect.
(defun html-protect-buffer ()
  (interactive)
  (goto-char (point-min))
  (replace-string "&"  "&amp;")

  (goto-char (point-min))
  (replace-string "<" "&lt;")

  (goto-char (point-min))
  (replace-string ">" "&gt;")

)



(defun make-laml-manual (dir0 title kind)
  "Make a LAML manual in dir0 which shows the photos located in this directory."
  (interactive "DMake a manual in which directory: 
sName of manual (without extension): 
nKind of manual: (1) SchemeDoc, (2) XML DTD, (3) Augmented XML DTD, or (4) custom manual: 
")
  (let ((dir (ensure-trailing-slash (expand-file-name dir0)))
       )
     (if (not (file-exists-p (concat dir title "." "laml")))
        (progn
            (let* ((manual-buf nil))

              ; Make the setup file
              (setq manual-buf
                    (make-a-file-from-laml-template 
                     (concat title "." "laml")
                     dir
                     (cond ((= kind 1) "manual-xml-in-laml-schemedoc")
                           ((= kind 2) "manual-xml-in-laml-xml-dtd")
                           ((= kind 3) "manual-xml-in-laml-xml-dtd-merge")
                           (t            "manual-xml-in-laml"))
                     'laml-mode
                      (list
                      )
                     ))

              (show-buffer (selected-window) manual-buf)
              (set-buffer-file-coding-system (quote iso-latin-1))

              (message "Done.  Fill in details and then LAML process the buffer.")

              ))

        (progn
	  (beep)
	  (message "The manual already exists. Nothing done.")))))